from typing import Optional

import pydantic
from pydantic.fields import Undefined

from ._gql_type import GQLType
from ._gql_variables import GQLVariables
from .serializers import class_to_graphql, value_to_graphql


class GQLQuery(GQLType):
    @classmethod
    def get_query_string(cls, include_name: bool = True) -> str:
        variables: Optional[
            type[GQLVariables]
        ] = cls.__config__.variables  # type: ignore

        def _generate():
            variables_str = _get_variables_string(variables)
            query_name = " " + cls.__name__ if include_name else ""
            yield "query" + query_name + variables_str + " {"
            for line in cls.generate_query_lines():
                yield "  " + line
            yield "}"

        return "\n".join(_generate())

    class Config(pydantic.BaseConfig):
        variables: Optional[type[GQLVariables]] = None


def _get_variables_string(variables: Optional[type[GQLVariables]]) -> str:
    if variables is None or not variables.__fields__:
        return ""

    def _generate():
        for field in variables.__fields__.values():
            type_str = class_to_graphql(field.type_, allow_none=field.allow_none)
            if field.allow_none and field.field_info.default is not Undefined:
                default_str = " = " + value_to_graphql(field.default)
            else:
                default_str = ""
            yield "$" + field.alias + ": " + type_str + default_str

    return "(" + ", ".join(_generate()) + ")"



import inspect
import typing
from types import UnionType
from typing import Any, Iterator

import pydantic

from .exceptions import QueryGenerationError
from .serializers import key_to_graphql, value_to_graphql


class GQLType(pydantic.BaseModel):
    @classmethod
    def generate_query_lines(cls) -> Iterator[str]:
        fields = typing.get_type_hints(cls)
        for field_name, field_type in fields.items():
            field = cls.__fields__[field_name]
            arguments_str = _get_arguments_string(field.field_info.extra)
            if typing.get_origin(field_type) is list:
                args = typing.get_args(field_type)
                if len(args) != 1:
                    raise QueryGenerationError(
                        f"Type '{field_type}' has unexpected amount of arguments"
                    )
                field_type = args[0]
            if typing.get_origin(field_type) is UnionType:
                sub_types = typing.get_args(field_type)
                yield field.alias + arguments_str + " {"
                for sub_type in sub_types:
                    if sub_type is object:
                        continue
                    if not issubclass(sub_type, GQLType):
                        raise QueryGenerationError(
                            f"Member '{sub_type}' of a union type"
                            f"must be a subtype of '{GQLType.__name__}'"
                        )
                    yield "  " + "... on " + sub_type.__name__ + " {"
                    for line in sub_type.generate_query_lines():
                        yield "  " * 2 + line
                    yield "  " + "}"
                yield "}"
                continue
            if not inspect.isclass(field_type):
                raise QueryGenerationError(f"Type {field_type} is not supported")
            if issubclass(field_type, GQLType):
                field_type.update_forward_refs()
                yield field.alias + arguments_str + " {"
                for line in field_type.generate_query_lines():
                    yield "  " + line
                yield "}"
                continue
            yield field.alias + arguments_str
            continue

    class Config:
        alias_generator = key_to_graphql
        allow_population_by_field_name = True


def _get_arguments_string(arguments: dict[str, Any]) -> str:
    if not arguments:
        return ""

    def _generate():
        for name, value in arguments.items():
            yield key_to_graphql(name) + ": " + value_to_graphql(value)

    return "(" + ", ".join(_generate()) + ")"



from typing import Any

import pydantic
import pydantic.main
from pydantic.fields import Field, FieldInfo
from pydantic.main import __dataclass_transform__

from .serializers import key_to_graphql


@__dataclass_transform__(kw_only_default=True, field_descriptors=(Field, FieldInfo))
class ModelMetaclass(pydantic.main.ModelMetaclass):
    def __getattr__(cls, __name: str) -> Any:
        try:
            mcs: type[GQLVariables] = cls  # type: ignore
            return mcs.__fields__[__name]
        except KeyError:
            raise AttributeError(
                f"type object '{cls.__name__}' has no attribute '{__name}'"
            )


class GQLVariables(pydantic.BaseModel, metaclass=ModelMetaclass):
    def _get_unset_defaults(self) -> set[str]:
        unset_defaults = set[str]()
        for field_name, field in self.__fields__.items():
            if (field_name in self.__fields_set__) or (not field.allow_none):
                continue
            unset_defaults.add(field_name)
        return unset_defaults

    def dict(
        self,
        by_alias: bool = True,
        exclude_defaults: bool = True,
        **kwargs: Any,
    ) -> dict[str, Any]:
        exclude = self._get_unset_defaults() if exclude_defaults else set()
        return super().dict(by_alias=by_alias, exclude=exclude, **kwargs)

    def json(
        self,
        by_alias: bool = True,
        exclude_defaults: bool = True,
        **kwargs: Any,
    ) -> str:
        exclude = self._get_unset_defaults() if exclude_defaults else set()
        return super().json(by_alias=by_alias, exclude=exclude, **kwargs)

    class Config:
        alias_generator = key_to_graphql
        allow_population_by_field_name = True



import json
from enum import Enum
from typing import Any

import inflection
from pydantic.fields import ModelField


_class_mapping: dict[type, str] = {
    int: "Int",
    float: "Float",
    str: "String",
    bool: "Boolean",
}


def register_graphql_type(graphql_type: str, python_class: type) -> None:
    _class_mapping[python_class] = graphql_type


def class_to_graphql(cls: type, allow_none: bool) -> str:
    suffix = "" if allow_none else "!"
    if issubclass(cls, Enum):
        return cls.__name__ + suffix
    try:
        type_ = _class_mapping[cls]
        return type_ + suffix
    except KeyError:
        raise TypeError(
            f"Type '{cls}' could not be converted to a GraphQL type."
            "See pygraphic.types.register_graphql_type"
        )


def value_to_graphql(value: Any) -> str:
    if type(value) is ModelField:
        return "$" + value.alias
    if isinstance(value, Enum):
        return value.name
    return json.dumps(value, indent=None, default=str)


def key_to_graphql(key: str) -> str:
    return inflection.camelize(key, uppercase_first_letter=False)



from . import exceptions, serializers
from ._gql_enum import GQLEnum
from ._gql_query import GQLQuery
from ._gql_type import GQLType
from ._gql_variables import GQLVariables


__all__ = [
    "exceptions",
    "GQLEnum",
    "GQLVariables",
    "GQLType",
    "GQLQuery",
    "serializers",
]



from enum import Enum
from typing import Any


class GQLEnum(Enum):
    @classmethod
    def validate(cls, value: Any) -> "GQLEnum":
        if isinstance(value, str):
            # Find Enum member by name
            try:
                return cls[value]
            except KeyError:
                pass
        return cls(value)

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def __modify_schema__(cls, field_schema: dict):
        field_schema.update(enum=[member.name for member in cls])



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pygraphic']

package_data = \
{'': ['*']}

install_requires = \
['inflection>=0.5.1,<0.6.0', 'pydantic>=1.9.1,<2.0.0']

setup_kwargs = {
    'name': 'pygraphic',
    'version': '0.5',
    'description': 'Client-side GraphQL query generator based on Pydantic',
    'long_description': '# pygraphic\n<p>\n<a href="https://pypi.org/project/pygraphic" target="_blank">\n    <img src="https://img.shields.io/pypi/status/pygraphic" alt="Status">\n</a>\n<a href="https://pypi.org/project/pygraphic" target="_blank">\n    <img src="https://img.shields.io/pypi/v/pygraphic" alt="Version">\n</a>\n<a href="https://github.com/lonelyteapot/pygraphic/actions/workflows/test.yml?query=branch%3Amain" target="_blank">\n    <img src="https://img.shields.io/github/workflow/status/lonelyteapot/pygraphic/Unit%20tests/main?label=tests" alt="Tests">\n</a>\n<a href="https://codecov.io/gh/lonelyteapot/pygraphic" target="_blank">\n    <img src="https://img.shields.io/codecov/c/github/lonelyteapot/pygraphic" alt="Coverage">\n</a>\n</p>\n\nClient-side GraphQL query generator based on [pydantic].\n\n## Why?\n\nWorking with GraphQL in Python seems simple... If you\'re fine with dictionaries, lack of\nautocompletion and unexpected errors.\n\nSome tools allow you to generate Python code from GraphQL schemas. One of them, [turms],\neven generates pydantic models from GQL documents. This approach can be problematic:\nqueries are written in GraphQL, not Python, so the codebase you\'re actually working with\nis out of your control; and the main advantage of pydantic — data validation — is\nmissing!\n\n## Workflow\n\nPygraphic is the opposite of [turms]:\n\n1. For each individual query, you define pydantic models that you want to request,\n   optionally with validators and other configuration;\n\n2. Pygraphic converts those definitions to raw GraphQL documents *(basically strings)*;\n\n3. Using a GraphQL or an HTTP client, you make requests with those documents and get\n   back JSON responses;\n\n4. Pydantic converts those responses to instances of the defined models and validates\n   them;\n\n5. You use the validated data, while enjoying autocompletion and type safety!\n\n## Roadmap\n\nPygraphic is in development stage. Some major features are missing or might work\nincorrectly. The API may change at any time.\n\nSee [ROADMAP.md][Roadmap] for the list of implemented/missing features.\n\n## Examples\n\nExamples are kept in the\n[/examples](https://github.com/lonelyteapot/pygraphic/tree/main/examples) folder.  \nQueries that they\'re expected to produce are stored in the\n[/golden_files](https://github.com/lonelyteapot/pygraphic/tree/main/folden_files)\nfolder.\n\nMost examples are based on [GitHub\'s GraphQL API](https://docs.github.com/en/graphql).\nIf you want to run them, you need to \n[generate a personal access token](https://github.com/settings/tokens) and assign it to\nthe environment variable called `GITHUB_TOKEN` (with VSCode, the best way to do this is\nto create a [`.env` file](https://www.dotenv.org/env) in the project\'s root directory).\n\nHere\'s the gist:\n\n### main.py\n\n``` python\nimport os\nimport requests\nfrom pygraphic import GQLQuery, GQLType\n\n# Define data models\nclass License(GQLType):\n    id: str\n    name: str\n    description: str\n\n# Define query model\nclass GetAllLicenses(GQLQuery):\n    licenses: list[License]\n\n# Generate the GraphQL query string\nquery_str = GetAllLicenses.get_query_string()\n\n# Make the request\nTOKEN = os.environ["GITHUB_TOKEN"]\nresponse = requests.post(\n    url="https://api.github.com/graphql",\n    json={\n        "query": query_str,\n    },\n    headers={\n        "Authorization": f"bearer {TOKEN}",\n    },\n)\n\n# Parse the response\nresponse_data = response.json().get(\'data\')\nif response_data is None:\n    print(\'Query failed\')\nresult = GetAllLicenses.parse_obj(response_data)\n\n# Print the results\nfor license in result.licenses:\n    print(license.name)\n```\n\n### Generated query string\n\n``` gql\nquery GetAllLicenses {\n  licenses {\n    id\n    name\n    description\n  }\n}\n```\n\n## Contribution\n\nThis project is developed on [GitHub].\n\nIf you have any general questions or need help — you\'re welcome in the [Discussions]\nsection.\n\nIf you encounter any bugs or missing features — file new [Issues], but make sure to\ncheck the existing ones first.\n\nIf you want to solve an issue, go ahead and create a [Pull Request][Pulls]! It will be\nreviewed and hopefully merged. Help is always appreciated.\n\n## License\n\nCopyright &copy; 2022, Dmitry Semenov. Released under the [MIT license][License].\n\n\n[GitHub]: https://github.com/lonelyteapot/pygraphic\n[Discussions]: https://github.com/lonelyteapot/pygraphic/discussions\n[Issues]: https://github.com/lonelyteapot/pygraphic/issues\n[Pulls]: https://github.com/lonelyteapot/pygraphic/pulls\n[License]: https://github.com/lonelyteapot/pygraphic/blob/main/LICENSE\n[Roadmap]: https://github.com/lonelyteapot/pygraphic/blob/main/ROADMAP.md\n\n[pydantic]: https://pypi.org/project/pydantic/\n[turms]: https://pypi.org/project/turms/\n',
    'author': 'Dmitry Semenov',
    'author_email': 'lonelyteapot@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/lonelyteapot/pygraphic',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
